@@ -27,7 +27,7 @@ yields!(
 
 yields!(
     rec_update,
-    "def upto($x): .[$x], (if $x > 0 then upto($x-1) else empty end); [1, 2, 3, 4] | upto(1) |= .+1",
+    "def upto($x): .[$x], (if $x > 0 then upto($x-1) else {}[] as $x | . end); [1, 2, 3, 4] | upto(1) |= .+1",
     [2, 3, 3, 4]
 );
 
@@ -72,59 +72,25 @@ yields!(first_empty, "[first({}[])]", json!([]));
 yields!(first_some, "first(1, 2, 3)", 1);
 
 yields!(
-    format_text,
-    r#"[0, 0 == 0, {}.a, "hello", {}, [] | @text]"#,
-    ["0", "true", "null", "hello", "{}", "[]"]
-);
-yields!(
-    format_json,
-    r#"[0, 0 == 0, {}.a, "hello", {}, [] | @json]"#,
-    ["0", "true", "null", "\"hello\"", "{}", "[]"]
-);
-yields!(
-    format_html,
-    r#""<p style='visibility: hidden'>sneaky</p>" | @html"#,
-    "&lt;p style=&apos;visibility: hidden&apos;&gt;sneaky&lt;/p&gt;"
-);
-yields!(
-    format_uri,
-    r#""abc123 ?#+&[]" | @uri"#,
-    "abc123%20%3F%23%2B%26%5B%5D"
-);
-yields!(
-    format_csv,
-    r#"[0, 0 == 0, {}.a, "hello \"quotes\" and, commas"] | @csv"#,
-    r#"0,true,,"hello ""quotes"" and, commas""#
-);
-yields!(
-    format_tsv,
-    r#"[0, 0 == 0, {}.a, "hello \"quotes\" and \n\r\t\\ escapes"] | @tsv"#,
-    "0\ttrue\t\thello \"quotes\" and \\n\\r\\t\\\\ escapes"
-);
-yields!(
-    format_base64,
-    r#""hello cruel world" | @base64"#,
+    encode_base64,
+    r#""hello cruel world" | encode_base64"#,
     "aGVsbG8gY3J1ZWwgd29ybGQ="
 );
 yields!(
-    format_unformat_base64,
-    r#""hello cruel world" | @base64 | @base64d"#,
+    encode_decode_base64,
+    r#""hello cruel world" | encode_base64 | decode_base64"#,
     "hello cruel world"
 );
+
 yields!(
-    format_sh,
-    r#"[0, 0 == 0, {}.a, "O'Hara!", ["Here", "there"] | @sh]"#,
-    ["0", "true", "null", r#"'O'\''Hara!'"#, r#"'Here' 'there'"#,]
+    escape_html,
+    r#""<p style='visibility: hidden'>sneaky</p>" | escape_html"#,
+    "&lt;p style=&apos;visibility: hidden&apos;&gt;sneaky&lt;/p&gt;"
 );
 yields!(
-    format_sh_rejects_objects,
-    r#"{a: "b"} | try @sh catch -1"#,
-    -1
-);
-yields!(
-    format_sh_rejects_nested_arrays,
-    r#"["fine, but", []] | try @sh catch -1"#,
-    -1
+    encode_uri,
+    r#""abc123 ?#+&[]" | encode_uri"#,
+    "abc123%20%3F%23%2B%26%5B%5D"
 );
 
 #[test]
@@ -220,20 +186,6 @@ fn limit() {
     give(json!(null), "[limit(-1; 0, 1)]", json!([]));
 }
 
-yields!(min_empty, "[] | min_by(.)", json!(null));
-// when output is equal, min_by selects the left element and max_by the right one
-yields!(
-    min_max_eq,
-    "[{a: 1, b: 3}, {a: 1, b: 2}] | [(min_by(.a), max_by(.a)) | .b]",
-    [3, 2]
-);
-// multiple-output functions can be used to differentiate elements
-yields!(
-    max_mult,
-    "[{a: 1, b: 3}, {a: 1, b: 2}] | max_by(.a, .b) | .b",
-    3
-);
-
 yields!(
     math_0_argument_scalar_filters,
     "[-2.2, -1.1, 0, 1.1, 2.2 | sin as $s | cos as $c | $s * $s + $c * $c]",
@@ -314,7 +266,7 @@ fn round() {
     give(json!(-1.4), "floor", json!(-2));
     give(json!(-1.4), "ceil", json!(-1));
 
-    let err = |v| Error::Type(Val::from(v), Type::Num);
+    let err = |v| Error::Type(Val::from(v), Type::Float);
     fail(json!([]), "round", err(json!([])));
     fail(json!({}), "round", err(json!({})));
 }
